/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt;
package data;

import scala.concurrent.FutureTaskRunner;
import scala.concurrent.ops.future;
import scalala.generic.CanAddInto;

object Concurrent {
  val numCores = Runtime.getRuntime.availableProcessors;

  /**
   * Sum a series of items in parallel into the heap parent of each item.
   * At return, values.head will contain the sum of all elements in the
   * map.
   */
  def treeReduce[V,C]
  (values : Seq[C], reduce : (C,C) => C,
   numThreads : Int = numCores)
  (implicit ft : FutureTaskRunner = new scala.concurrent.ThreadRunner()) : C = {

    if (values.length == 1) {
      return values(0);
    }

    val valueQueue = new scala.collection.mutable.HashMap[Int,C]();
    def setValue(i : Int, value : C) = valueQueue.synchronized {
      require(!valueQueue.contains(i));
      valueQueue(i) = value;
    }
    def getValue(i : Int) = valueQueue.synchronized {
      valueQueue.remove(i) match {
        case Some(v) => v;
        case None => values(i);
      }
    }

    def parent(i : Int) = (i-1)/2;

    def left(i : Int) = (i*2)+1;
    def hasLeft(i : Int) = left(i) < values.size;

    def right(i : Int) = (i*2)+2;
    def hasRight(i : Int) = right(i) < values.size;

    def doTopDown(numHeads : Int) : Int = {
      if (numHeads == 0) {
        return values.size;
      }

      // heap property: take heads from frstHead to lastHead (inclusive)
      val lastHead = (2 * (numHeads-1));
      val frstHead = lastHead - numHeads + 1;

      def processTree(i : Int) : Unit = {
        val l = left(i);
        if (l < values.size) {
          processTree(l);
          setValue(i, reduce(getValue(i),getValue(l)));
        }
        val r = right(i);
        if (r < values.size) {
          processTree(r);
          setValue(i, reduce(getValue(i),getValue(r)));
        }
      }

      // System.err.println("[TreeSum] TopDown("+numHeads+") does "+frstHead+" .. "+lastHead);

      (frstHead to lastHead).toList.map(i => future(processTree(i))(ft)).map(_.apply())
      lastHead+1;
    }

    def doBottomUp(upto : Int) = {
      // how many signals must we wait for at each position: 0, 1, or 2?
      val latches = Array.tabulate(upto)(i =>
        new java.util.concurrent.CountDownLatch(
          (if (left(i) < upto) 1 else 0) +
          (if (right(i) < upto) 1 else 0)));

      def process(i : Int) : Unit = {
        latches(i).await;
        if (i > 0) {
          val p = parent(i);
          latches(p) synchronized {
            setValue(p, reduce(getValue(p),getValue(i)));
          }
          latches(p).countDown;
        }
      }

      val active = new java.util.concurrent.atomic.AtomicInteger(upto);

      def compute() {
        while (true) {
          val next = active.decrementAndGet;
          if (next >= 0) {
            process(next);
          } else {
            return;
          }
        }
      }

      // System.err.println("[TreeSum] BottomUp("+upto+")");

      List.tabulate(numThreads)(i => future(compute())(ft)).map(_.apply());
    }

    doBottomUp(doTopDown(math.min(values.size / (2*numThreads), numThreads)));
    require(valueQueue.size == 1);
    valueQueue(0);
    //doTopDown(1);
    //doBottomUp(values.size);
  }

  /**
   * Sum a series of items in parallel into the heap parent of each item.
   * At return, values.head will contain the sum of all elements in the
   * map.
   */
  def treeSum[V,C](values : Seq[C], numThreads : Int = numCores)
  (implicit sum : CanAddInto[C,C], ft : FutureTaskRunner) =
    treeReduce(values, ((a:C,b:C) => { sum(a,b); a; }), numThreads);

  /**
   * Use numCores cores to map the given iterator.  The values are returned
   * in the same order as the original iterator, but each item is mapped in
   * its own thread.
   */
  def map[A,B](iterator : Iterator[A], fn : (A=>B), numCores : Int = numCores) : Iterator[B] = {
    val iHasNext = Array.fill(numCores)(new java.util.concurrent.ArrayBlockingQueue[Boolean](1));
    val iNext    = Array.fill(numCores)(new java.util.concurrent.ArrayBlockingQueue[A](1));
    val oHasNext = Array.fill(numCores)(new java.util.concurrent.ArrayBlockingQueue[Boolean](1));
    val oNext    = Array.fill(numCores)(new java.util.concurrent.ArrayBlockingQueue[B](1));
    
    val threads = new scala.collection.mutable.ArrayBuffer[Thread] with
      scala.collection.mutable.SynchronizedBuffer[Thread];

    val exceptions = new scala.collection.mutable.ArrayBuffer[Throwable] with
      scala.collection.mutable.SynchronizedBuffer[Throwable];

    // spawn: send each data item to a channel
    scala.concurrent.ops.spawn {
      try {
        threads += Thread.currentThread;
        var i = 0;
        for (item <- iterator) {
          iHasNext(i % numCores).put(true);
          iNext(i % numCores).put(item);
          i += 1;
        }
        iHasNext.foreach(_.put(false));
      } catch {
        case ie : InterruptedException =>
          // done
        case ex : Throwable =>
          exceptions += ex;
      }
    }

    // spawn: consume data items from channels
    for (shard <- 0 until numCores) {
      scala.concurrent.ops.spawn {
        try {
          threads += Thread.currentThread;
          while (iHasNext(shard).take) {
            oHasNext(shard).put(true);
            oNext(shard).put(fn(iNext(shard).take));
          }
          oHasNext(shard).put(false);
        } catch {
          case ie : InterruptedException =>
            // done
          case ex : Throwable =>
            exceptions += ex;
        }
      }
    }

    // return: iterator from reading channels
    new Iterator[B] {
      var i = 0;
      var knowsHasNext = false;
      var valofHasNext = false;

      override def hasNext = {
        if (!exceptions.isEmpty) {
          threads.foreach(_.interrupt());
          throw exceptions(0);
        }

        if (!knowsHasNext) {
          knowsHasNext = true;
          valofHasNext = oHasNext(i % numCores).take;
          if (!valofHasNext) {
            // consume remaining (false) hasNexts
            val ok = oHasNext.drop((i % numCores) + 1).forall(_.take == false);
            require(ok, "Some iterators not done?");
          }
        }
        valofHasNext;
      }

      override def next = {
        if (!hasNext) {
          throw new NoSuchElementException();
        }
        val rv = oNext(i % numCores).take;
        knowsHasNext = false;
        i += 1;
        rv;
      }

      override protected def finalize = {
        threads.foreach(_.interrupt());
      }
    }
  }

  def mapreduce[T,V:ClassManifest](
    items : TraversableOnce[T], map : (T=>V), reduce : ((V,V) => V),
    numCores : Int = numCores)
  : V = {
    val partials = new Array[V](numCores);

    val threads = new scala.collection.mutable.ArrayBuffer[Thread] with
      scala.collection.mutable.SynchronizedBuffer[Thread];

    val exceptions = new scala.collection.mutable.ArrayBuffer[Throwable] with
      scala.collection.mutable.SynchronizedBuffer[Throwable];

    val queue = new java.util.concurrent.ArrayBlockingQueue[T](numCores * 2);

    val atomic = new java.util.concurrent.atomic.AtomicInteger(0);

    // spawn processor threads
    (0 until numCores).foreach(core => future {
      threads += Thread.currentThread;
      try {
        partials(core) = map(queue.take());
        atomic.decrementAndGet;

        while (true) {
          val item = queue.take();
          partials(core) = reduce(partials(core),map(item));
          atomic.decrementAndGet;
        }
      } catch {
        case ex : Throwable =>
          exceptions += ex;
      }
    });

    // supply values
    for (item <- items) {
      if (!exceptions.isEmpty) {
        threads.foreach(_.interrupt());
        throw exceptions.head;
      }

      atomic.incrementAndGet;
      queue.put(item);
    }

    while (atomic.get > 0) {
      Thread.`yield`;
    }
    threads.foreach(_.interrupt());

    treeReduce(partials, reduce);
  }

  def main(args : Array[String]) {
    implicit val ft = new scala.concurrent.ThreadRunner();
    for (threads <- List(1,2,4,8,16,32,64); size <- List(10,100,1000,10000,100000,1000000,10000000)) {
      println("[TreeSum] 0 until "+size+" using "+threads+" threads");
      val value = Array.tabulate(size)(i => Array(i));
      treeSum(value, threads);
      println("  "+value(0)(0)+" vs "+(0 until size).sum);
    }
  }
}
