/*
 * Distributed as part of the Stanford Topic Modeling Toolbox.
 * Copyright (c) 2009- The Board of Trustees of the Leland
 * Stanford Junior University.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA,
 * 02110-1301, USA.
 */
package edu.stanford.nlp.tmt;
package data;

import scalanlp.collection.LazyIterable;
import scalanlp.text.tokenize.Tokenizer;
import scalanlp.util.Index;

//package newversion {
//
//  trait Attribute[Item,+V] extends (Item=>V);
//
//  trait Identifier[Item,ID] extends Attribute[Item,ID] {
//    val id = this;
//  }
//
//  trait Dataset[Item,ID,This] {
//    def name : String;
//    def items : Iterable[Item];
//    def id : Identifier[Item,ID];
//  }
//
//  /**
//   * An (inner) joined view of two datasets.
//   *
//   * @author dramage
//   */
//  class JoinedDataset[ItemA,DatasetA,ItemB,DatasetB,ID](val _1 : DatasetA, val _2 : DatasetB)
//  extends Dataset[(ItemA,ItemB),ID,JoinedDataset[ItemA,DatasetA,ItemB,DatasetB,ID]] {
//
//    private val c1 = _1.asInstanceOf[Dataset[ItemA,ID,_]];
//    private val c2 = _2.asInstanceOf[Dataset[ItemB,ID,_]];
//
//    override def name = "Join("+c1.name+"," +c2.name+")";
//
//    override def items =
//      data.Dataset.join(c1.items, c1.id, c2.items, c2.id);
//
//    val id = new Identifier[(ItemA,ItemB),ID] {
//      def apply(tup : (ItemA,ItemB)) = c1.id(tup._1);
//    }
//  }
//
//  object JoinedDataset {
//    implicit def unrollLeft[DatasetA](joined : JoinedDataset[_,DatasetA,_,_,_])
//    : DatasetA = joined._1;
//
//    implicit def unrollRight[DatasetB](joined : JoinedDataset[_,_,_,DatasetB,_])
//    : DatasetB = joined._2;
//  }
//
//  object Dataset {
//    implicit def iDatasetToIdentifier[Item,ID,_](dataset : Dataset[Item,ID,_]) =
//      dataset.id;
//  }
//
//}

/**
 * A Dataset is a collection of Items with a function that can return
 * a unique identifier for each item.
 *
 * @author dramage
 */
trait Dataset[Item,ID,This] {
  /** Name of this dataset. */
  def name : String;

  /** Collection of items. */
  def items : Iterable[Item];

  /** Returns a unique identifier for each item. */
  def getID(item : Item) : ID;

  /** A view of this dataset as its underlying terms. */
  def asIDs =
    items.map(item => getID(item));

  def size =
    items.size;

  /** Signature is the signature of the name. */
  def signature =
    scalanlp.ra.Signature(name);

//  /** Joins the two datasets on their ids. */
//  def join[ItemB,That](that : That) =
//    new JoinedDataset[Item,This,ItemB,That,ID](this.asInstanceOf[This], that);

  /** Returns this.name. */
  override def toString = name;
}

///**
// * An (inner) joined view of two datasets.
// *
// * @author dramage
// */
//class JoinedDataset[ItemA,DatasetA,ItemB,DatasetB,ID](val _1 : DatasetA, val _2 : DatasetB)
//extends Dataset[(ItemA,ItemB),ID,JoinedDataset[ItemA,DatasetA,ItemB,DatasetB,ID]] {
//
//  private val c1 = _1.asInstanceOf[Dataset[ItemA,ID,_]];
//  private val c2 = _2.asInstanceOf[Dataset[ItemB,ID,_]];
//
//  override def name = "Join("+c1.name+"," +c2.name+")";
//
//  override def items =
//    Dataset.join(c1.items, c1.getID, c2.items, c2.getID);
//
//  override def getID(item : (ItemA,ItemB)) =
//    c1.getID(item._1);
//}
//
///**
// * Implicitly unroll a JoinedDataset into its left and right components.
// *
// * @author dramage
// */
//object JoinedDataset {
//  implicit def unrollLeft[DatasetA](joined : JoinedDataset[_,DatasetA,_,_,_])
//  : DatasetA = joined._1;
//
//  implicit def unrollRight[DatasetB](joined : JoinedDataset[_,_,_,DatasetB,_])
//  : DatasetB = joined._2;
//}

object Dataset {
  protected class QueuedIterator[T,ID](iter : Iterator[T], id : T=>ID) extends Iterator[T] {
    val queue = new scala.collection.mutable.Queue[T]();
    val idSet = new scala.collection.mutable.HashSet[ID]();

    override def hasNext =
      !queue.isEmpty || iter.hasNext;

    override def next = {
      if (!queue.isEmpty) {
        val item = queue.dequeue;
        idSet -= id(item);
        item;
      } else {
        iter.next;
      }
    }

    def enqueue : ID = {
      val item = iter.next;
      val itemId = id(item);
      queue.enqueue(item);
      require(idSet.add(itemId), "Duplicate id "+itemId+" in iterator");
      itemId;
    }

    /** Drops elements from the queue until the given id. */
    def discardUntil(to : ID) = {
      require(idSet.contains(to), "Can only discard up to items that have been queued.");
      var itemId = id(queue.head);
      while (itemId != to) {
        idSet.remove(itemId);
        queue.dequeue;
        itemId = id(queue.head);
      }
    }

    /** Returns true if the given id has already been queued. */
    def hasQueuedID(id : ID) =
      idSet.contains(id);

    /** Returns true if underlying iterator still has remaining items. */
    def hasUnqueued =
      iter.hasNext;
  }

  /**
   * Performs an inner join on the two iterables using the given id
   * functions to get their join id's.
   */
  def join[A,B,ID]
  (a : Iterable[A], idA : A=>ID, b : Iterable[B], idB : B => ID)
  : LazyIterable[(A,B)] =
    join(a,idA,b,idB,(ia:A,ib:B) => (ia,ib));

  /**
   * Performs an inner join on the two iterables using the given id
   * functions to get their join id's.
   */
  def join[A,B,C,ID]
  (a : Iterable[A], idA : A=>ID, b : Iterable[B], idB : B=> ID, fn : (A,B) => C)
  : LazyIterable[C] = new LazyIterable[C] {
    override def iterator = new Iterator[C] {
      val aIter = new QueuedIterator(a.iterator, idA);
      val bIter = new QueuedIterator(b.iterator, idB);

      private val LEFT = true;
      private val RIGHT = false;

      // initial state depends on next status
      var current : Option[C] = prepare();

      var pull = LEFT;

      /** Move until the next matching pair. */
      def prepare() : Option[C] = {
        while (aIter.hasUnqueued || bIter.hasUnqueued) {
          if (pull == LEFT) {
            if (aIter.hasUnqueued) {
              val id = aIter.enqueue;
              if (bIter.hasQueuedID(id)) {
                aIter.discardUntil(id);
                bIter.discardUntil(id);
                return Some(fn(aIter.next, bIter.next));
              }
            }
            pull = RIGHT;
          } else if (pull == RIGHT) {
            if (bIter.hasUnqueued) {
              val id = bIter.enqueue;
              if (aIter.hasQueuedID(id)) {
                bIter.discardUntil(id);
                aIter.discardUntil(id);
                return Some(fn(aIter.next, bIter.next));
              }
            }
            pull = LEFT;
          }
        }
        return None;
      }

      override def hasNext =
        current.isDefined;

      override def next = {
        val rv = current.get;
        current = prepare();
        rv;
      }
    }
  }
}


/**
 * A TermDataset has a set of terms associated with each item, and, optionally
 * an index of terms and a tokenizer describing how those terms were created.
 *
 * @author dramage
 */
trait TermDataset[Item] {
  this : Dataset[Item,_,_] =>

  /** The tokenizer used to tokenize this dataset. */
  def getTerms(item : Item) : Iterable[String];

  /** A view of this dataset as its underlying terms. */
  def asTerms(termIndex : Index[String] = this.termIndex.get) =
    items.map(item => getTerms(item).view.map(termIndex).filter(_ >= 0).toArray);

  //
  // Optional attributes
  //

  /** The index of terms in this dataset. */
  val termIndex : Option[Index[String]] = None;

  /** The base tokenizer used when generating this dataset. */
  val tokenizer : Option[Tokenizer] = None;
}


/**
 * A MultiLabelDataset has a set of labels associated with each item, and,
 * optionally an index of labels.
 *
 * @author dramage
 */
trait MultiLabelDataset[Item,Label] {
  this : Dataset[Item,_,_] =>

  /** The tokenizer used to tokenize this dataset. */
  def getLabels(item : Item) : Iterable[Label];

  /** A view of this dataset as an Iterable of Labels. */
  def asLabels(labelIndex : Index[Label] = this.labelIndex.get) =
    items.map(item => getLabels(item).view.map(labelIndex).filter(_ >= 0).toArray);

  /** Optional index of labels in this dataset. */
  val labelIndex : Option[Index[Label]] = None;
}
